// $Id: ta-util.hh 8460 2024-10-20 17:36:45Z flater $

/*
    TA (Transparent Archivist) (C) 2005 David Flater.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#include <Dstr>
#include <exception>
#include <vector>
#include <mhash.h>

// mhash.h clobbers VERSION and PACKAGE_VERSION.  The original value of
// VERSION as set by the configure script for TA gets stored here.
extern char const * const ta_version;

// Type to store sizes in bytes.
typedef unsigned long long int ta_t;

// "Standard" optical media sizes in bytes.  (1 sector = 2048 B)

// CD capacities and their mandatory pre- and post-gaps are
// well-specified, as are the capacities of +R and +RW media.  The
// capacity of DVD+R DL appears to be standardized, though I have
// found no explicit statement of this on the net.

// The capacities of dash-R media are not well-specified.  The one
// DVD-R that I tested (CMC MAG. AE1) had a capacity of 2297888
// sectors, which disagrees with the de facto standard alleged by OSTA
// (2298496).  I have never even seen a DVD-R DL.

// I have found nothing on the net about any pre- or post-gap on DVDs,
// but Nero 6 very specifically demands an extra 150 sectors.

// 5/9 are single/double layer (per the DVD-5 and DVD-9 monikers).

// The value of BDRsize was determined by testing to failure with Verbatim
// media and growisofs on Linux.  It is not necessarily standard.

#define Gapsize          307200ULL
#define CD74size    ( 681984000ULL - Gapsize * 2)
#define CD80size    ( 737280000ULL - Gapsize * 2)
#define DVD5size    (4700372992ULL - Gapsize    )
#define DVD9size    (8547991552ULL - Gapsize    )
#define BDRsize     24756879360ULL

// Stuff about files.
struct FileStuff {
  Dstr name;
  ta_t size;
  Dstr hash; // Stringified SHA-512, null for symlinks
};

// Type to store the big file list.
typedef std::vector<FileStuff> FileList;
typedef unsigned long FileListIndex;

// Some exceptions...
class FileReadException: public std::exception {};
class FileWriteException: public std::exception {};

// in hash_ID:  one of MHASH_SHA512, MHASH_MD5, etc.
//   (see list in /usr/include/mutils/mglobal.h)
// in filename:  name of a regular file (not a directory, not a symlink).
// out shash:  stringified hash of that file
//   (e.g., for SHA-512 it would be 128 characters).
// Throws FileReadException if there's a problem reading the file,
//   whether because it doesn't exist or whatever.
void getHash (hashid hash_ID, const Dstr &filename, Dstr &shash);

// Execute an external command.
//
// outfname    Name of file to which to redirect stdout (optional).
//
// argv[0]     The command
// argv[1]...  Arguments
// argv[n]     The list of arguments MUST be terminated by a NULL pointer.
//
// If the command returns a nonzero exit status, this is thrown as an
// exception of type int.
//
// All other types of failures result in exit with a helpful error message.
void sys (char const * const outfname, char const * const argv[]);
void sys (char const * const argv[]);

// Create a file of the specified size.
// If the file exists it is overwritten.
// All errors result in exit.
//
// If fill_random, file is filled with random data.
// Otherwise, it is created sparse if possible.
void createFile (const Dstr &filename, ta_t size, bool fill_random);

// Given:  original filename (regular file, directory or symlink) and wdpath.
// Effect:  delete the file from the working directory (NOT THE ORIGINAL).
// If the file does not exist, soldier on.  All other errors result in
// exit.
void rm (const Dstr &filename, const Dstr &wdpath);

// Make sure that a directory name is nonnull and ends in a slash.
void slashterm (Dstr &dirname);

// Given:  filename.
// Return:  name of corresponding file in wdpath.
void getTarget (const Dstr &filename, const Dstr &wdpath, Dstr &target);

// Given:  a file or directory name.
// Effect:  add that file or recursively add contents of that directory to
//          flist.
//          . and .. references are fixed only if fixDots is true.
//          If compressSizes, size is set based on zlib estimate.
// Directories themselves are not put on the list.
// Directories need not have trailing slashes.
void addSomething (const Dstr &fname, FileList &flist, bool fixDots,
		   bool compressSizes);

// Attempt to write hashes to workingdir/ta-hashes.txt.
// In case of failure, hash file is removed and FileWriteException is thrown.
// Written hashes are flist[0] through flist[ptr-1].
void write_hashes (const Dstr &workingdir, const FileList &flist,
		   FileListIndex ptr);

// Given:  file or directory name.
// Effect:  change the modes of directories and executable regular files to
//          0755 and of non-executable regular files to 0644.
// (ta does not use this on directories but does them separately)
void setperm_file (const Dstr &target);

// Call before exit to close libmagic nicely.
void setperm_cleanup ();
